#pragma implementation

#include "AuxMethodSignature.h"
#include "Srcstream.h"
#include "JavaMethodSignature.h"
#include "find_leaks.h"
#include <stdio.h>

#define MAX_INT_STR_SIZE 64

//
// Method name : (constuctor)
// Description : Creates a new AuxMethodSignature, representing the
//    "null AuxMethod".
CAuxMethodSignature::CAuxMethodSignature()
  : fType(UNK),
    fClass(CJavaTypeSignature::kVoid),
    fWType(Virtual),
    fMethod(0),
    fFirstMethod(0),
    fThrows(0),
    fName(0),
    fWidth(0),
    fDim(0),
    fFirstClass(0)
{  
  REF_PTR(this);
}

//
// Method name : (copy constructor)
//
CAuxMethodSignature::CAuxMethodSignature(const CAuxMethodSignature& other)
  : fType(other.fType),
    fClass(other.fClass),
    fWType(other.fWType),
    fMethod(0),
    fFirstMethod(0),
    fThrows(0),
    fName(0),
    fWidth(other.fWidth),
    fDim(other.fDim),
    fFirstClass(0)
{
  REF_PTR(this);
  if (other.fMethod)
    fMethod = new CJavaMethodSignature(*other.fMethod);
  if (other.fFirstMethod)
    fFirstMethod = new CJavaMethodSignature(*other.fFirstMethod);
  if (other.fThrows)
    fThrows = new deque<CJavaTypeSignature>(*other.fThrows);
  if (other.fName) {
    fName = new unicode_string(*other.fName);
  }
  if (other.fFirstClass)
    fFirstClass = new CJavaTypeSignature(*other.fFirstClass);
  assert(repOK());
}

//
// Method name : (constructor)
// Description : Creates an AuxMethod of type "type", associated with the 
//     class "Class".  Other fields must be supplied where necessary.
//   * All arguments are adopted, not copied.
//   * If firstMethod and firstClass are not provided (default to 0), the
//     corresponding fields default to the proper values.
//
CAuxMethodSignature::CAuxMethodSignature(AuxMethodType type, 
				     const CJavaTypeSignature &Class,
				     unicode_string *name,
				     CJavaMethodSignature *method,
				     deque<CJavaTypeSignature> *throws,
				     WhereType wtype,
				     unsigned width, unsigned dim, 
				     CJavaMethodSignature *firstMethod,
				     CJavaTypeSignature *firstClass)
  : fType(type),
    fClass(Class),
    fWType(wtype),
    fMethod(method),
    fFirstMethod(firstMethod),
    fThrows(throws),
    fName(name),
    fWidth(width),
    fDim(dim),
    fFirstClass(firstClass)

{
  REF_PTR(this);
  if (fMethod != 0 && fFirstMethod == 0) {
    fFirstMethod = new CJavaMethodSignature(*fMethod);
  }
  if ((fType == SetStatic || fType == FetchStatic || fType == Cast) 
      && !fFirstClass) {
    fFirstClass = new CJavaTypeSignature(fClass);
  }
  assert(repOK());
}    

//
// Destructor
//
CAuxMethodSignature::~CAuxMethodSignature()
{
  delete fName;
  delete fMethod;
  delete fFirstMethod;
  delete fThrows;
  delete fFirstClass;
  UNREF_PTR(this);
}

//
// Method name : compare
// Description : Return <0 if this is less than other, 0 if this == other,
//     and >0 if this is greater than other.
//
int
CAuxMethodSignature::compare(const CAuxMethodSignature& other) const
{
  assert(repOK() && other.repOK());
  int result = other.fType - fType;
  if (result != 0) return result;
  result = fClass.Compare(other.fClass);
  if (result != 0) return result;
  switch(fType) {
  case Where:
    result = other.fWType - fWType;
    if (result != 0) return result;
    result = fName->compare(*other.fName);
    if (result != 0) return result;
    result = fMethod->Compare(*other.fMethod);
    if (result != 0) return result;
    result = fFirstMethod->Compare(*other.fFirstMethod);
    if (result != 0) return result;
    break;
  case ArrayNew:
    result = other.fWidth - fWidth;
    if (result != 0) return result;
    result = other.fDim - fDim;
    if (result != 0) return result;
    break;
  case SetStatic:
  case FetchStatic:
    result = fName->compare(*other.fName);
    if (result != 0) return result;
    result = fFirstClass->Compare(*other.fFirstClass);
    if (result != 0) return result;
    break;
  case Cast:
    result = fFirstClass->Compare(*other.fFirstClass);
    if (result != 0) return result;
    break;    
  default:
    break;
  }
  return 0;
}

//
//  Method name: ASCCompile
//  Description: Return the representation of this AuxMethodSignature,
//      as written to the .pjaux file.
//
string
CAuxMethodSignature::ASCCompile() const
{
  string s;
  assert(repOK());
  switch (fType)
    {
    case Where:
      s = "Where";
      break;
    case ArrayStore:
      s = "ArrayStore";
      break;
    case ArrayFetch:
      s = "ArrayFetch";
      break;
    case ArrayLength:
      s = "ArrayLength";
      break;
    case ArrayNew:
      s = "ArrayNew";
      break;
    case GetWhere: 
      s = "GetWhere";
      break;
    case SetStatic:
      s = "SetStatic";
      break;
    case FetchStatic:
      s = "FetchStatic";
      break;
    case PSig:
      s = "PSig";
      break;
    case Cast:
      s = "Cast";
      break;  
    case InstanceOf:
      s = "InstanceOf";
      break;  
    case Default:
      s = "Default";
      break;
    default:
      assert(0);
    }
    s += ":";
    s += fClass.Compile();
  
    char buf[MAX_INT_STR_SIZE];

    switch (fType) {
    case Where:
      s += (fWType == Constructor) ? ":C:" :
	((fWType == Virtual) ? ":V:" : ":S:");
      s += fMethod->CompileType();
      s += ":";
      s += fFirstMethod->CompileType();
      s += ":";
      for(deque<CJavaTypeSignature>::iterator i = fThrows->begin();
	  i != fThrows->end(); ++i) {
	s += (*i).Compile();
      }
      s += ":";
      s += ::UnicodeToUTF(*fName);
      break;
    case ArrayNew:
      s += ":";
      sprintf(buf, "%d", fWidth);
      s += buf;
      s += ":";
      sprintf(buf, "%d", fDim);
      s += buf;
      break;
    case SetStatic:
    case FetchStatic:
      s += ":";
      s += fFirstClass->Compile();      
      s += ":";
      s += ::UnicodeToUTF(*fName);      
      break;
    case Cast:
      s += ":";
      s += fFirstClass->Compile();
      break;
    default:
      break;
    }
    return s;
}

//
// Method name : IsInstantiated
// Description : Returns true iff this AuxMethod should be instantiated with
//    its current signature (i.e., its deferred information is now resolved,
//    and there are no unbound parameters in fClass of fMethod).
//
bool
CAuxMethodSignature::IsInstantiated() const
{
  assert(fType != UNK);
  return fClass.IsInstantiable() && 
    ((fMethod == 0) || fMethod->IsInstantiable());
}

//
// Method name : Unparse
// Description : Writes this auxmethod to sOut, implementing it if possible.
//    Assumes that this auxmethod has been assigned the name "name". Registers
//    instantiated types on the instantiationList.
//
void
CAuxMethodSignature::Unparse(const unicode_string& name,
			     srcstream& sOut,
			     TypeSet& instantiationList,
			     CTypeResolver& tRes,
			     const CJavaTypeSignature &inType,
			     bool headerOnly) const
{
  assert(repOK());

  bool fieldProtected = false;
  if (fType == FetchStatic || fType == SetStatic) {
    bool ret = tRes.IsVarProtected(inType, *fName, fieldProtected);
    assert(ret);
  }

  ////
  // Header
  ////

  //   Flags  
  sOut.out() << "public ";
  if (! IsInstantiated() || headerOnly) 
    sOut.out() << "abstract ";
  
  //   Return type
  switch(fType)
    {
    case SetStatic:  
      if (fFirstClass->IsPrimitive())
	fFirstClass->Unparse(sOut);
      else 
	sOut.out() << "Object";
      break;
    case ArrayStore:
      sOut.out() << "Object";
      break;
    case ArrayFetch:
      sOut.out() << "Object";
      break;
    case ArrayNew:
      sOut.out() << "Object";
      for(int i = fDim-1; i > 0; --i) {
	sOut.out() << "[]";
      }
      break;
    case ArrayLength:
      sOut.out() << "int";
      break;
    case Where:
      if (fWType == Constructor)
	sOut.out() << "Object";
      else
	fFirstMethod->GetType().Unparse(sOut);
      break;
    case GetWhere:
      fClass.UnparseWType(sOut);
      break;
    case FetchStatic:
      if (fFirstClass->IsPrimitive())
	fFirstClass->Unparse(sOut);
      else 
	sOut.out() << "Object";      
      break;
    case InstanceOf:
      sOut.out() << "boolean";
      break;
    case Cast:
      fFirstClass->Unparse(sOut);
      break;
    case PSig:
      sOut.out() << "String";
      break;
    case Default:
      sOut.out() << "Object";
      break;      
    default:
      assert(0);
    }
  sOut.out() << " ";

  //   Name
  sOut.out() << ::UnicodeToUTF(name)
	     << "(";
  
  //   Arguments
  switch(fType)
    {
    case GetWhere:
    case PSig:
    case Default:
      break;
    case SetStatic:
      if (fFirstClass->IsPrimitive())
	fFirstClass->Unparse(sOut);
      else 
	sOut.out() << "Object";
      sOut.out() << " $val";
      if (fieldProtected) {
	sOut.out() << ", ";
	inType.UnparseTokenClass(sOut);
	sOut.out() << " $t";
      }
      break;
    case FetchStatic:
      if (fieldProtected) {
	inType.UnparseTokenClass(sOut);
	sOut.out() << " $t";
      }
      break;
    case ArrayStore:
      sOut.out() << "Object array, int index, Object val";
      break;
    case ArrayFetch:
      sOut.out() << "Object array, int index";
      break;
    case ArrayLength:
      sOut.out() << "Object array";
      break;
    case ArrayNew:
      for (unsigned int i = 1; i <= fWidth; i++) {
	sOut.out() << "int $"
		   << i;
	if (i < fWidth)
	  sOut.out() << ", ";
      }
      break;
    case InstanceOf:
      sOut.out() << "Object val";
      break;
    case Where:
      {
	if (fWType == Virtual) {
	  sOut.out() << "Object $this";
	  if (fFirstMethod->ParameterCount() > 0)
	    sOut.out() << ", ";
	}
      	int n = 1;
	for (deque<CJavaTypeSignature>::const_iterator i = 
	       fFirstMethod->ParametersBegin();
	     i != fFirstMethod->ParametersEnd(); ++n) {
	  (*i).Unparse(sOut);
	  sOut.out() << " $"
		     << n;
	  ++i;
	  if (i != fFirstMethod->ParametersEnd())
	    sOut.out() << ", ";
	}
      }
    break;
    case Cast:
      sOut.out() << "Object val";
      break;
    default:
      assert(0);
    }
  sOut.out() << ")";
  // Exceptions
  if (fType == Where) {
    assert(fThrows);
    if (fThrows->size() >0) {
      sOut.out() << " throws ";
      for(deque<CJavaTypeSignature>::iterator excpt = fThrows->begin();
	  excpt != fThrows->end(); ) {
	(*excpt).Unparse(sOut);
	if (++excpt != fThrows->end()) sOut.out() << ", ";
      }
    }
  }

  // Do we implement it or not?
  if (!IsInstantiated() || headerOnly) {
    sOut.out() << ";";
    sOut.Next();
    return;
  } else {
    sOut.PushIndent(2).Next().out() << "{ ";
  }
     
  ////
  // Implementation
  ////
  switch(fType)
    {
    case GetWhere:
      {
	unicode_string fill;
	sOut.out() << "return ";
	fClass.UnparseInstantiation(sOut,tRes);
	sOut.out() << ".getInstance();";
	instantiationList.insert(fClass);
      }
      break;
    case ArrayStore:
      {
	sOut.out() << "((";
	fClass.Unparse(sOut);
	sOut.out() << "[]) array)[index] = ";
	if (fClass.IsPrimitive()) {
	  sOut.out() << "((" << ::UnicodeToUTF(fClass.GetWrapperName())
		     << ") val).value;";
	} else {
	  sOut.out() << "(";
	  fClass.Unparse(sOut);
	  sOut.out() << ") val;";
	}
	sOut.Next().out() << "return val;";
      }
      break;
    case ArrayFetch:
      {
	if (fClass.IsPrimitive()) {
	  sOut.out() << "return ";
	  if (fClass.GetBaseType() == CJavaTypeSignature::Boolean) {
	    sOut.out() << ::UnicodeToUTF(fClass.GetWrapperName())
		       << ".Wrap(((";
	    fClass.Unparse(sOut);
	    sOut.out() << "[]) array)[index]);";
	  } else {
	    sOut.out() << "(Object) new "
		       <<::UnicodeToUTF(fClass.GetWrapperName())
		       << "(((";
	    fClass.Unparse(sOut);
	    sOut.out() << "[]) array)[index]);";
	  }
	} else {
	  sOut.out() << "return ((";
	  fClass.Unparse(sOut);
	  sOut.out() << "[]) array)[index];";
	}
      }
      break;
    case ArrayLength:
      {
	sOut.out() << "return ((";
	fClass.Unparse(sOut);
	sOut.out() << "[]) array).length;";
      }
      break;
    case ArrayNew:
      {
	sOut.out() << "return (Object";
	for(unsigned int i = 0; i < fDim-1; ++i) 
	  sOut.out() << "[]";
	sOut.out() << ") new ";
	fClass.Unparse(sOut);
	for(unsigned int i = 1; i <= fWidth; ++i) 
	  sOut.out() << "[$" << i << "]";
	for(unsigned int i = 1; i <= fDim-fWidth; ++i) 
	  sOut.out() << "[]";
        sOut.out() << ";";
      }
    break;
    case FetchStatic: 
      {
	if (fieldProtected) {
	  sOut.out() 
	    << "if ($t == null) throw new java.lang.SecurityException();";
	  sOut.Next();
	}
	sOut.out() << "return "
		   << ::SymbolToUTF(*fName)
		   << ";";
      }
      break;
    case SetStatic:
      {
	if (fieldProtected) {
	  sOut.out() 
	    << "if ($t == null) throw new java.lang.SecurityException();";
	  sOut.Next();
	}
	sOut.out() << ::SymbolToUTF(*fName)
		   << " = $val;";
	sOut.Next().out() << "return $val;";
      }
      break;      
    case InstanceOf:
      {
	// We only need to call the "$instanceof" method if we have a
        // parameterized type.  Otherwise we can use the "instanceof"
	// operator
	if (fClass.IsParameterized()) {
	  sOut.out() << "return ";
	  fClass.UnparseWType(sOut);
	  sOut.out() << ".$instanceof((Object) val";  
	  for(deque<CJavaTypeSignature>::const_iterator i = 
		fClass.GetParameters()->begin();
	      i != fClass.GetParameters()->end(); ++i) {
	    sOut.out() << ", \"" << (*i).Compile() << "\"";
	  }
	  sOut.out() << ");";
	  instantiationList.insert(fClass);
	} else {
	  sOut.out() << "return (val instanceof ";
	  if (fClass.IsPrimitive()) {
	    sOut.out() << ::UnicodeToUTF(fClass.GetWrapperName());
	  } else {
	    fClass.Unparse(sOut);
	  }
	  sOut.out() << ");";
	}
      }
      break;
    case Cast:
      {
	// We only need to call the "$cast" method if we have a
        // parameterized type.
	if (fClass.IsParameterized()) {
	  sOut.out() << "return ";
	  fClass.UnparseWType(sOut);
	  sOut.out() << ".$cast((Object) val";  
	  for(deque<CJavaTypeSignature>::const_iterator i = 
		fClass.GetParameters()->begin();
	      i != fClass.GetParameters()->end(); ++i) {
	    sOut.out() << ", \"" << (*i).Compile() << "\"";
	  }
	  sOut.out() << ");";
	  instantiationList.insert(fClass);
	} else if (fClass.IsPrimitive()) {	 
	  sOut.out() << "return "
		     << ::UnicodeToUTF(fClass.GetWrapperName())
		     << ".Cast(val);";
	} else {
	  sOut.out() << "return ((";
	  fClass.Unparse(sOut);
	  sOut.out() << ") val);";
	}
      }
      break;
    case PSig:
      {
	sOut.out() << "return \""
		   << fClass.Compile()
		   << "\";";
      }
      break;
    case Where:
      {
	if (fClass.IsPrimitive() || fClass.IsArray()) {
	  if (fWType == Constructor)
	    UnparsePrimitiveConstructor(sOut);
	  else
	    UnparsePrimitiveMethod(sOut);
	} else { 
	  CJavaMethodSignature declMethod;
	  CJavaMethodSignature transMethod;
	  bool res = tRes.GetDeclaredMethodType(fClass, *fMethod, 
						declMethod, transMethod);
	  bool retPrim = fFirstMethod->GetType().IsPrimitive();
	  bool havePrim = transMethod.GetType().IsPrimitive();
	  bool wrapRet = havePrim && !retPrim;
	  bool unWrapRet = retPrim && !havePrim;
	  assert(res);
	  if (fWType == Constructor ||
	      !(fMethod->GetType() == CJavaTypeSignature::kVoid))
	    sOut.out() << "return ";

	  if (fWType == Constructor) {
	    sOut.out() << "(Object) new ";
	    fClass.Unparse(sOut);
	    sOut.out() << "(";
	    UnparseArgsToCall(sOut,transMethod,tRes);
	    sOut.out() << ");";
	  } else {
	    if (wrapRet) {
	      sOut.out() << "new " 
			 << ::UnicodeToUTF(fMethod->GetType().GetWrapperName())
			 << "(";
	    }
	    if (fWType == Virtual) {
	      sOut.out() << "((";
	      fClass.Unparse(sOut);
	      sOut.out() << ") $this).";
	    } else {
	      fClass.Unparse(sOut);
	      sOut.out() << ".";
	    }
	    sOut.out() << ::SymbolToString(*fName)
		       << "(";
	    UnparseArgsToCall(sOut,transMethod,tRes);
	    if (unWrapRet) {
	      sOut.out() << ").value;";
	    } else if (wrapRet) {
	      sOut.out() << "));";
	    } else {
	      sOut.out() << ");";
	    }
	  }
	}
      }
      break;
    case Default:
      sOut.out() << "return ";
      if (! fClass.IsPrimitive()) 
	sOut.out() << "null;";
      else {
	sOut.out() << ::UnicodeToUTF(fClass.GetWrapperName())
		   << ".Default;";
      }
      break;
    default:
      assert(0);
    }
    sOut.out() << " }";
    sOut.PopIndent(2).Next();
}

//
//  Method name: UnparseArgsToCall
//  Description: Write the arguments to a call to a constraint method
//      represented by "fMethod" for the implementation of this AuxMethod,
//      casting them as needed.
//
void
CAuxMethodSignature::UnparseArgsToCall(
		            srcstream &sOut,
			    const CJavaMethodSignature& transMethod,
			    CTypeResolver &tRes) const
{
  assert(fMethod && fFirstMethod);
  deque<CJavaTypeSignature>::const_iterator iter =
    fMethod->ParametersBegin();
  deque<CJavaTypeSignature>::const_iterator end =
    fMethod->ParametersEnd();
  deque<CJavaTypeSignature>::const_iterator fmiter =
    fFirstMethod->ParametersBegin();
  deque<CJavaTypeSignature>::const_iterator tmiter
    = transMethod.ParametersBegin();
  
  ////
  // First, do we need a WObject?
  ////
  if (fWType != Virtual && fClass.IsParameterized()) {
    fClass.UnparseInstantiation(sOut,tRes);
    sOut.out() << ".getInstance()";
    if (iter != end)
      sOut.out() << ", ";
  }

  for (int n = 1; iter != end; ++fmiter, ++tmiter, ++n) {
    bool fmPrim = (*fmiter).IsPrimitive();
    bool tmPrim = (*tmiter).IsPrimitive();
    bool unwrapping = !fmPrim && tmPrim; // Do we unwrap our input?
    bool wrapping = !tmPrim && fmPrim; // Do we wrap our output?
    bool casting = !wrapping && !unwrapping;
    if (!tmPrim && !fmPrim && (*iter).IsPrimitive())
      casting = false;
    if (wrapping) {
      if ((*iter).GetBaseType() == CJavaTypeSignature::Boolean) {
	sOut.out() << ::UnicodeToUTF((*iter).GetWrapperName())
		   << ".Wrap(";
      } else {
	sOut.out() << "(Object) new "
		   << ::UnicodeToUTF((*iter).GetWrapperName())
		   << "(";
      }
    }
    if (unwrapping) {
      sOut.out() << "((" << ::UnicodeToUTF((*iter).GetWrapperName())
		 << ") ";
    }
    if (casting) {
      sOut.out() << "(";
      (*iter).Unparse(sOut);
      sOut.out() << ") ";
    }    
    sOut.out() << "$"
	       << n;
    if (unwrapping) sOut.out() << ").value";
    if (wrapping) sOut.out() << ")";
    ++iter;
    if (iter != end) {
      sOut.out() << ", ";
    }
  }
}

//
//  Method name : UnparsePrimitiveMethod
//  Description : Write the body of this AuxMethod implementation (minus the
//      brackets) to sOut.
//  Checks      : This is a valid primitive method.
//
void
CAuxMethodSignature::UnparsePrimitiveMethod(srcstream& sOut) const
{
  assert(fType == CAuxMethodSignature::Where);
  assert(fWType == CAuxMethodSignature::Virtual);
  assert(fClass.IsPrimitive() || fClass.IsArray());
  assert(fFirstMethod);
  assert(fName);

  ////
  // Commonly used strings.
  ////
  // Do we return a primitive type?
  bool returnPrim = fFirstMethod->GetType().IsPrimitive();
  // The unwrapped value of 'this'.
  string uThis;
  if (fClass.IsPrimitive())
    uThis = "((" + ::UnicodeToUTF(fClass.GetWrapperName()) + ") $this).value";
  // The unwrapped value of the first argument.
  string uFirst;
  if (fFirstMethod->ParameterCount() >0) {
    uFirst = UnparseUnwrapArg("$1", 
			      *fFirstMethod->ParametersBegin(),
			      *fMethod->ParametersBegin());
  }
  // How do we 'wrap' a primitive value for output?
  string wrapRet;
  string wrapRetClose = ")";
  if (returnPrim) {
    wrapRet = "(" +::SymbolToString(fFirstMethod->GetType().Disassemble()) 
      +")(";
  } else if (fMethod->GetType() == CJavaTypeSignature::kBoolean) {
    wrapRet = ::UnicodeToUTF(fMethod->GetType().GetWrapperName()) +
      ".Wrap("; 
  } else if (fMethod->GetType().IsPrimitive()) {    
    wrapRet = "new " + ::UnicodeToUTF(fMethod->GetType().GetWrapperName())
      + "((" + ::SymbolToString(fMethod->GetType().Disassemble()) + ")( ";
    wrapRetClose = "))";
  } else {
    CJavaTypeSignature t(fFirstMethod->GetType());
    t.Objectify();
    wrapRet = "(" + ::SymbolToString(t.Disassemble()) + ")(";
  }
  #define WRAP(x) (wrapRet + (x) + wrapRetClose)
  #define SWRAP(x) wrapRet << x << wrapRetClose
  // Zero.
  string zero = "((" + ::SymbolToString(fClass.Disassemble()) + ")0)";
  // How do we return 'this'?
  string retThis  = returnPrim ? WRAP(uThis) : string("$this");
  string retFirst;
  if (!returnPrim && (fFirstMethod->ParameterCount() != 0) &&
      !(*fFirstMethod->ParametersBegin()).IsPrimitive() &&
      (*fMethod->ParametersBegin()) == fMethod->GetType()) {
    retFirst = "$1";
  } else {
    retFirst = WRAP(uFirst);
  }

  ////
  // Implementations
  ////
  string name = ::UnicodeToUTF(*fName);
  if (name == "abs") {
    assert(fClass.IsNumeric());
    if (fClass.GetBaseType() == CJavaTypeSignature::Character) {
      sOut.out() << "return $this;";
    } else {
      // "return ($this >= 0) ? $this : (-$this);"
      string negationString = "-(" + uThis + ")";
      sOut.out() << "return ("
		 << uThis
		 << " >= " << zero 
		 << ") ? " << WRAP(uThis) << " : " 
		 << WRAP(negationString) << ";";
    }
  } else if (name == "add") {
    assert(fClass.IsNumeric());
    sOut.out() << "return " << SWRAP(uThis << "+" << uFirst) << ";";
  } else if (name == "and") {
    assert(fClass.IsOrdinal() || 
	   fClass.GetBaseType() == CJavaTypeSignature::Boolean);
    sOut.out() << "return " << SWRAP(uThis
	       << (fClass.IsOrdinal() ? "&" : "&&")
	       << uFirst) << ";";
  } else if (name == "booleanValue") {
    assert(fClass.GetBaseType() == CJavaTypeSignature::Boolean);
    sOut.out() << "return " << retThis << ";";
  } else if (name == "byteValue") {
    assert(fClass.IsNumeric());
    sOut.out() << "return " << retThis << ";";
  } else if (name == "charValue") {
    sOut.out() << "return " << retThis << ";";
  } else if (name == "clone" || name == "$clone_array") {
    if (fClass.IsPrimitive()) {
      // All primitive types are immutable, so we can return this object.
      sOut.out() << "return " << retThis << ";";
    } else {
      // Clone the array:
      // elems[][] a = (elems[][])$this;
      // Object x = (Object) new elems[a.length][];
      // System.arraycopy($this, 0, x, 0, a.length);
      // return x;
      CJavaTypeSignature el(fClass);
      el.SetArrayBounds(0);
      fClass.Unparse(sOut);
      sOut.out() << " a = (";
      fClass.Unparse(sOut);
      sOut.out() << ") $this;";
      sOut.PushIndent(2).Next().out() << "Object x = (Object) new ";
      el.Unparse(sOut);
      sOut.out() << "[a.length]";      
      for(unsigned int k = 1; k < fClass.GetArrayBounds(); ++k) 
	sOut.out() << "[]";
      sOut.out() << ";";
      sOut.Next().out()
	<< "System.arraycopy((Object)$this, 0, x, 0, a.length);";
      sOut.Next().PopIndent(2).out() << "return x;";
#if 0
      // XXXX This won't work under Java 1.0.
     
      // Clone the array:
      // return ((elems[][])$this).clone();
      sOut.out() << "return ((";
      fClass.Unparse(sOut);
      sOut.out() << ")$this).clone();";
#endif
    }
  } else if (name == "compareTo") {
    // "int x = $this; int y = $1; return (x < y) ? -1 : ((x == y) ? 0 : 1);"
    assert(fClass.IsNumeric());
    assert(fFirstMethod && fFirstMethod->ParameterCount() == 1);
    assert(fMethod      && fMethod->ParameterCount()      == 1);
    bool unwrapArg = ! (*(fFirstMethod->ParametersBegin())).IsPrimitive();
    string arg = unwrapArg ? "y" : "$1";
    fClass.Unparse(sOut);
    sOut.out() << " x = " << uThis << "; ";
    if (unwrapArg) {
      (*(fMethod->ParametersBegin())).Unparse(sOut);
      sOut.out() << " y = " << uFirst << "; "; 
    }
    sOut.out() << "return (x < " << arg << ") ? -1 : ((x == " << arg
	       << ") ? 0 : 1);";
  } else if (name == "divide") {
    assert(fClass.IsNumeric());
    sOut.out() << "return " << SWRAP(uThis << "/" << uFirst) << ";";
  } else if (name == "doubleValue") {
    sOut.out() << "return " << retThis << ";";
  } else if (name == "elementAt" || name == "get") {
    assert(fClass.IsArray());
    CJavaTypeSignature baseType(fClass);
    baseType.SetArrayBounds(baseType.GetArrayBounds()-1);
    sOut.out() << "return " << wrapRet << "((";
    fClass.Unparse(sOut);
    sOut.out() << ") $this)[$1]" << wrapRetClose << ";";
  } else if (name == "equals") {
    assert(fFirstMethod && fFirstMethod->ParameterCount() == 1);
    assert(fMethod      && fMethod->ParameterCount()    == 1);
    CJavaTypeSignature   argType = *fMethod->ParametersBegin();
    if (fClass.IsArray()) {
      sOut.out() << "return $this.equals($1);";
    } else if (argType.IsReference()) {
      unicode_string argName;
      assert(argType.GetBaseClassName(argName) && 
	     argName == CJavaTypeSignature::kJLOName);
      sOut.out() << "return polyj.lang.Primitive.ee($this, $1);";
    } else {
      sOut.out() << "return " << uThis << " == " << uFirst << ";";
    }
  } else if (name == "less") {
    sOut.out() << "return " << uThis << " < " << uFirst << ";";
  } else if (name == "greater") {
    sOut.out() << "return " << uThis << " > " << uFirst << ";";
  } else if (name == "floatValue") {
    sOut.out() << "return " << retThis << ";";
  } else if (name == "hashCode") {
    sOut.out() << "return $this.hashCode();";
  } else if (name == "intValue") {
    sOut.out() << "return " << retThis << ";";
  } else if (name == "size") {
    assert(fClass.IsArray());
    sOut.out() << "return ((";
    fClass.Unparse(sOut);
    sOut.out() << ") $this).length;";
  } else if (name == "longValue") {
    sOut.out() << "return " << retThis << ";";
  } else if (name == "max") {
    // "return ($this > $1) ? $this : $1;
    assert(fClass.IsNumeric());
    assert(fFirstMethod && fFirstMethod->ParameterCount() == 1);
    assert(fMethod      && fMethod->ParameterCount()      == 1);
    sOut.out() << "return (" << uThis << " > " << uFirst << ") ? "
	       << retThis << " : " << retFirst << ";";
  } else if (name == "min") {
    // "return ($this < $1) ? $this : $1;
    assert(fClass.IsNumeric());
    assert(fFirstMethod && fFirstMethod->ParameterCount() == 1);
    assert(fMethod      && fMethod->ParameterCount()      == 1);
    sOut.out() << "return (" << uThis << " < " << uFirst << ") ? "
	       << retThis << " : " << retFirst << ";";
  } else if (name == "mod") {
    assert(fClass.IsOrdinal());
    sOut.out() << "return " << SWRAP(uThis << "%" << uFirst) << ";";
  } else if (name == "multiply") {
    assert(fClass.IsNumeric());
    sOut.out() << "return " << SWRAP(uThis << "*" << uFirst) << ";";
  } else if (name == "negate") {
    assert(fClass.IsNumeric());
    sOut.out() << "return " << SWRAP("-(" << uThis << ")") << ";";
  } else if (name == "not") {
    assert(fClass.IsOrdinal() ||
 	   fClass.GetBaseType() == CJavaTypeSignature::Boolean);
    string notString = (fClass.IsOrdinal() ? "~" : "!");
    sOut.out() << "return " << SWRAP(notString << "(" << uThis << ")") << ";";
  } else if (name == "or") {
    assert(fClass.IsOrdinal() ||
	   fClass.GetBaseType() == CJavaTypeSignature::Boolean);
    sOut.out() << "return " << SWRAP(uThis
	       << (fClass.IsOrdinal() ? "|" : "||")
	       << uFirst) << ";";
  } else if (name == "setElementAt") { 
    assert(fClass.IsArray());
    assert(fFirstMethod && fFirstMethod->ParameterCount() == 2);
    assert(fMethod      && fMethod->ParameterCount()    == 2);
    string uSecond = UnparseUnwrapArg("$2", 
				      *(++fFirstMethod->ParametersBegin()),
				      *(++fMethod->ParametersBegin()));
    sOut.out() << "((";
    fClass.Unparse(sOut);
    sOut.out() << ") $this)[" << uSecond << "] = "
	       << uFirst << ";";
  } else if (name == "set") {
    assert(fClass.IsArray());
    assert(fFirstMethod && fFirstMethod->ParameterCount() == 2);
    assert(fMethod      && fMethod->ParameterCount()    == 2);
    string uSecond = UnparseUnwrapArg("$2", 
				      *(++fFirstMethod->ParametersBegin()),
				      *(++fMethod->ParametersBegin()));
    sOut.out() << "((";
    fClass.Unparse(sOut);
    sOut.out() << ") $this)[" << uFirst << "] = "
	       << uSecond << ";";
  } else if (name == "shortValue") {
    sOut.out() << "return " << retThis << ";";
  } else if (name == "signum") {
    // "int x = $this; return (x < 0) ? -1 : ((x==0)?0:1);"
    assert(fClass.IsNumeric());
    fClass.Unparse(sOut);
    sOut.out() << " x = " << uThis
	       << "; return (x < " << zero << ") ? -1 : ((x == "
	       << zero << ") ? 0 : 1);";
  } else if (name == "subtract") {
    assert(fClass.IsNumeric());
    sOut.out() << "return " << SWRAP(uThis << "-" << uFirst) << ";";
  } else if (name == "toString") {
    sOut.out() << "return $this.toString();";
  } else if (name == "xor") {
    assert(fClass.IsOrdinal() || 
	   fClass.GetBaseType() == CJavaTypeSignature::Boolean);
    sOut.out() << "return " << SWRAP(uThis << "^" << uFirst) << ";";
  } else {
    // No such method!
    assert(0);
  }  
}

//
//  Method name : UnparsePrimitiveConstructor
//  Description : Write the body of this AuxMethod implementation (minus the
//      brackets) to sOut.
//  Checks      : This is a valid primitive constructor.
//
void
CAuxMethodSignature::UnparsePrimitiveConstructor(srcstream& sOut) const
{
  assert(fType == CAuxMethodSignature::Where);
  assert(fWType == Constructor);
  assert(fClass.IsPrimitive() || fClass.IsArray());
  assert(fName && (::UnicodeToUTF(*fName) == "<init>"));
  assert(fMethod && fFirstMethod);

  if (fClass.IsArray()) {
    // Array constructor.
    assert(fFirstMethod->ParameterCount() == 1);
    assert(fMethod->ParameterCount()      == 1);

    const CJavaTypeSignature& argType = *(fMethod->ParametersBegin());
    if (argType.IsArray()) {
      // Clone the array:
      // elems[][] a = (elems[][])$1;
      // Object x = (Object) new elems[a.length][];
      // System.arraycopy($this, 0, x, 0, a.length);
      // return x;
      CJavaTypeSignature el(fClass);
      el.SetArrayBounds(0);
      fClass.Unparse(sOut);
      sOut.out() << " a = (";
      fClass.Unparse(sOut);
      sOut.out() << ") $1;";
      sOut.PushIndent(2).Next().out() << "Object x = (Object) new ";
      el.Unparse(sOut);
      sOut.out() << "[a.length]";      
      for(unsigned int k = 1; k < fClass.GetArrayBounds(); ++k) 
	sOut.out() << "[]";
      sOut.out() << ";";
      sOut.Next().out() << "System.arraycopy((Object)$1, 0, x, 0, a.length);";
      sOut.Next().PopIndent(2).out() << "return x;";
    } else {
      assert(argType.GetBaseType() == CJavaTypeSignature::Integer);
      CJavaTypeSignature member = fClass;
      member.SetArrayBounds(member.GetArrayBounds()-1);
      sOut.out() << "return new ";
      member.Unparse(sOut);
      sOut.out() << "[$1];";
    }
  } else {
    if (fMethod->ParameterCount() == 0) {
      sOut.out() << "return "
		 << ::UnicodeToUTF(fClass.GetWrapperName())
		 << ".Default;";
    } else {
      // Not an array array, Parameter count == 1;
      assert(fMethod->ParameterCount() == 1);
      string uFirst = UnparseUnwrapArg("$1", 
				       *fFirstMethod->ParametersBegin(),
				       *fMethod->ParametersBegin());
      if (fClass.GetBaseType() == CJavaTypeSignature::Boolean) {
	sOut.out() << "return "
		   << ::UnicodeToUTF(fClass.GetWrapperName())
		   << ".Wrap(" << uFirst << ");";
      } else {
	sOut.out() << "return new "
		   << ::UnicodeToUTF(fClass.GetWrapperName())
		   << "((";
	fClass.Unparse(sOut);
	sOut.out() << ") " << uFirst << ");";
      }
    }
  }
}

//
//  Method name : UnparseUnwrapArg
//  Description : Return the code necessary to unwrap the argument
//     value written in "unwrap", where objType is the objectified argument
//     type, and type is its actual type.
//
string
CAuxMethodSignature::UnparseUnwrapArg(const string& unwrap,
				      const CJavaTypeSignature& firstType,
				      const CJavaTypeSignature& type) const
{
  if (type.IsPrimitive() && ! firstType.IsPrimitive()) {
    string s;
    s += "((" + ::UnicodeToUTF(type.GetWrapperName()) +
      ") " + unwrap + ").value";
    return s;
  } else if (!(firstType == type)) {
    string s = "(" + ::SymbolToString(type.Disassemble()) + ")" + unwrap;
    return s;
  } else {
    return unwrap;
  }
}

//
// Method name: Disassemble
// Description: Dumps a description of this auxmethod to out.  For debugging
//    purposes.
//
void
CAuxMethodSignature::Disassemble(ostream &out) const
{
  switch(fType)
    {
    case CAuxMethodSignature::Where:
      out << "Where       ";
      break;
    case CAuxMethodSignature::ArrayStore:
      out << "ArrayStore  ";	
      break;     
    case CAuxMethodSignature::ArrayFetch:
      out << "ArrayFetch  ";
      break;     
    case CAuxMethodSignature::ArrayLength:
      out << "ArrayLength ";
      break;     
    case CAuxMethodSignature::ArrayNew:
      out << "ArrayNew    ";
      break;     
    case CAuxMethodSignature::GetWhere:
      out << "GetWhere    ";
      break;     
    case CAuxMethodSignature::SetStatic:
      out << "SetStatic   ";
      break;     
    case CAuxMethodSignature::FetchStatic:
      out << "FetchStatic ";	
      break;     
    case CAuxMethodSignature::InstanceOf:
      out << "InstanceOf  ";
      break;     
    case CAuxMethodSignature::Cast:
      out << "Cast        ";
      break;     
    case CAuxMethodSignature::PSig:
      out << "PSig        ";
      break;     
    case CAuxMethodSignature::Default:
      out << "Default     ";
      break;     
    default:
      out << "UNKNOWN:    " << fType;
    }
    out << "(Class: ";
    fClass.Disassemble(out);
    if (fName)
      out << ") (Name: " << ::UnicodeToUTF(*fName); 
    if (fMethod) 
      out << ") (Method: " << *fMethod;
    if (fFirstMethod)
      out << ") (FirstMethod: " << *fFirstMethod;
    if (fFirstClass)
      out << ") (FirstClass: " << *fFirstClass;
    if (fThrows) {
      out << ") (Throws:";
      for (deque<CJavaTypeSignature>::iterator i = fThrows->begin();
	   i != fThrows->end(); ++i)
	out << " " << (*i);
    }
    if (fType == Where) {
      out << ") (WType: ";
      if (fWType == Constructor) 
	out << "Constructor";
      else if (fWType == Virtual) 
	out << "Virtual";
      else if (fWType == Static) 
	out << "Static";
      else 
	out << "UNK";
    }
    if (fType == ArrayNew)
      out << ") (Width: " << fWidth << ") (Dimensions: " << fDim;
    out << ")";
}

bool 
CAuxMethodSignature::repOK() const
{
  switch (fType) {
  case Where:
    if (fWType != Constructor && fWType != Static && fWType != Virtual)
      return false;
    return (fMethod && fFirstMethod && fThrows && fName && !fFirstClass);
    break;
  case SetStatic:
  case FetchStatic:
    return (!fMethod && !fFirstMethod && !fThrows && fName && fFirstClass);
    break;
  case ArrayStore:
  case ArrayFetch:
  case ArrayLength:
  case ArrayNew:
  case GetWhere:
  case PSig:
  case InstanceOf:
  case Default:
    return (!fMethod && !fFirstMethod && !fThrows && !fName && !fFirstClass);
    break;
  case Cast:
    return (!fMethod && !fFirstMethod && !fThrows && !fName && fFirstClass);
    break;
  default:    
    return false;
  };
  assert(0);
}

ostream& operator<<(ostream &o, const CAuxMethodSignature &ams)
{
  ams.Disassemble(o);
  return o;
}
