// $Id: JavaMethodInfo.C,v 1.13 1998/08/20 18:28:12 nickm Exp $
// Copyright (c) 1995  David Engberg  All rights reserved
#pragma implementation
#include "find_leaks.h"
#include "JavaMethodInfo.h"
#include "JavaCodeAttribute.h"
#include "JavaExceptionsTable.h"
#include <iostream.h>

//
//  Method name : CJavaMethodInfo
//  Description : Default constructor.
//
CJavaMethodInfo::CJavaMethodInfo()
  : fCodeAttribute(0),
    fExceptions(0),
    fReferenceCount(0),
    fDeprecated(false),
    fThrows(0)
{
}

//
//  Method name : ~CJavaMethodInfo
//  Description : Destructor
//
CJavaMethodInfo::~CJavaMethodInfo()
{
  assert(fReferenceCount == 0);
  delete fCodeAttribute;
  delete fExceptions;
  delete fThrows;
}

//
//  Method name : ParseBuffer
//  Description : Parses the input buffer at the point specified and tries
//    to interpret it as a method information entry in a Java class file.
//    If successful, a new java method information node is returned,
//    otherwise, 0 is returned. If the 'interfaceOnly' flag is true,
//    then the method is loaded without any code attributes for the
//    methods.  This can be used whenever the goal is to get the signatures
//    of methods (like during compilation) without the implementations.
//
CJavaMethodInfo*
CJavaMethodInfo::ParseBuffer(string::const_iterator& javaBuffer,
			     const CJavaClassFile& classFile,
			     bool interfaceOnly)
{
  CJavaMethodInfo* result = new CJavaMethodInfo;
  result->fAccessFlags.SetFlags(CJavaClassFile::ReadJavaU2(javaBuffer));

  JavaConstantIndex nameIndex = CJavaClassFile::ReadJavaU2(javaBuffer);
  const CJavaAscizConstant* tempAscizConstant =
    DYNAMIC_CAST_CONST(CJavaAscizConstant, 
		       classFile.LookupConstant(nameIndex));
  if (tempAscizConstant == 0) {
    delete result; return 0;
  }
  unicode_string methodName = tempAscizConstant->GetUnicodeString();
  JavaConstantIndex signatureIndex = CJavaClassFile::ReadJavaU2(javaBuffer);
  tempAscizConstant = DYNAMIC_CAST_CONST(CJavaAscizConstant, 
		      classFile.LookupConstant(signatureIndex));
  if (tempAscizConstant == 0) {
    delete result; return 0;
  }
  unicode_string signatureString = tempAscizConstant->GetUnicodeString();
  if (! result->fSignature.Initialize(methodName, signatureString)) {
    delete result; return 0;
  }
  result->fTranslatedSignature = result->fSignature;

  unsigned long attributeCount = CJavaClassFile::ReadJavaU2(javaBuffer);
  while (attributeCount-- > 0) {
    CJavaAttribute* attribute =
      CJavaAttribute::ParseBuffer(javaBuffer, classFile);
    if (attribute != 0) {
      CJavaCodeAttribute* codeAttribute =
	DYNAMIC_CAST(CJavaCodeAttribute, attribute);
      if (codeAttribute != 0 && !interfaceOnly) {
	delete result->fCodeAttribute;
	result->fCodeAttribute = codeAttribute;
      } else {
	CJavaExceptionsTable* exceptions =
	  DYNAMIC_CAST(CJavaExceptionsTable, attribute);
	if (exceptions != 0) {
	  delete result->fExceptions;
	  result->fExceptions = exceptions;
	} else {
	  if (DYNAMIC_CAST(CJavaDeprecatedAttribute, attribute) != 0) {
	    result->fDeprecated = true;
	  }
	  delete attribute;
	}
      }
    }
  }
  return result;
}

//
//  Method name : Disassemble
//  Description : Dumps out a human-readable description of this method to
//    the provided stream.  For debugging purposes.
//
void
CJavaMethodInfo::Disassemble(ostream& toStream) const
{
  fSignature.Disassemble(toStream);
  toStream << " ";
  fAccessFlags.Disassemble(toStream);
  if (fDeprecated) {
    toStream << "  (Deprecated)";
  }
  if (fExceptions != 0) {
    toStream << endl << "    ";
    fExceptions->Disassemble(toStream);
  }
  if (fCodeAttribute != 0) {
    fCodeAttribute->Disassemble(toStream);
  }
}

//
//  Method name : AddReference
//  Description : Increments this object's reference count.
//
void
CJavaMethodInfo::AddReference() const
{
  // casting away const
  ((CJavaMethodInfo*)this)->fReferenceCount++;
}

//
//  Method name : RemoveReference
//  Description : Decrements this object's reference count.  Deletes the object
//    if the count drops to 0.
//
void
CJavaMethodInfo::RemoveReference() const
{
  assert(fReferenceCount > 0);
  // casting away const
  ((CJavaMethodInfo*)this)->fReferenceCount--;
  if (fReferenceCount == 0) {
    delete this;
  }
}

//
//  Method name : ThrowsExceptions
//  Description : This method returns true if this method throws any exceptions
//
bool
CJavaMethodInfo::ThrowsExceptions() const
{
  return fExceptions != 0 && fExceptions->Size() != 0;
}

//
//  Method name : ThrowsBegin
//  Description : This method returns an iterator over the exceptions thrown
//    by this method.
//
deque<unicode_string>::const_iterator
CJavaMethodInfo::ThrowsBegin() const
{
  if (fExceptions == 0) {
    // casting away const
    ((CJavaMethodInfo*)this)->fExceptions = new CJavaExceptionsTable;
  }
  return fExceptions->ThrowsBegin();
}

//
//  Method name : ThrowsEnd
//  Description : This returns an iterator to the end of the list of exceptions
//    thrown by this method.
//
deque<unicode_string>::const_iterator
CJavaMethodInfo::ThrowsEnd() const
{
  if (fExceptions == 0) {
    // casting away const
    ((CJavaMethodInfo*)this)->fExceptions = new CJavaExceptionsTable;
  }
  return fExceptions->ThrowsEnd();
}

//
//  Method name : GetThrows
//  Description : This method returns a reference to a deque of all the
//     exceptions thrown by this method.
const deque<CJavaTypeSignature> &
CJavaMethodInfo::GetThrows() const
{
  if (fThrows == 0) {    
    deque<CJavaTypeSignature> *throws = new deque<CJavaTypeSignature>;
    for (deque<unicode_string>::const_iterator i = ThrowsBegin();
	 i != ThrowsEnd(); ++i) {
      CJavaTypeSignature t(*i);
      throws->push_back(t);
    }
    // Cast away const
    ((CJavaMethodInfo*) this)->fThrows = throws;
  }
  return *fThrows;
}

//
//  Method name : IsDeprecated
//  Description : Returns true if this method is deprecated from use.
//
bool
CJavaMethodInfo::IsDeprecated() const
{
  return fDeprecated;
}

//
//  Method name : SetSignature
//  Description : Private method to revise the signature of a method.
//      Used by CPJavaClassFile::ParseFromString to denote a method's
//      true signature.
//
void 
CJavaMethodInfo::SetSignature(const CJavaMethodSignature& sig) const
{
  ((CJavaMethodInfo*)this)->fSignature = sig;
}

//
//  Method name : SetTranslatedSignature
//  Description : Private method to revise the signature of a method.
//      Used by CCompiler::RenamePJMethods to denote a method's
//      true signature.
//
void 
CJavaMethodInfo::SetTranslatedSignature(const CJavaMethodSignature& sig) const
{
  ((CJavaMethodInfo*)this)->fTranslatedSignature = sig;
}


//
//  Method name : GetTranslatedSignature
//  Description : Return the translated signature of this method.
//
const CJavaMethodSignature& 
CJavaMethodInfo::GetTranslatedSignature() const
{
  if (fTranslatedSignature.GetName().length() == 0) {
    return fSignature;
  }
  return fTranslatedSignature;
}
